package common.core;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.ServerSocket;
import java.net.UnknownHostException;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.UUID;

import org.apache.log4j.Logger;

import device.structures.Device;

/**
 *
 * <pre>
 *                 Copyright (c) 2006 Dominik Schulz
 *                 Copyright (c) 2006 Florian Lindner
 *                 Copyright (c) 2006 Philip Hartmann
 *
 *                 This file is part of jUPNP.
 *
 *                 jUPNP is free software; you can redistribute it and/or modify
 *                 it under the terms of the GNU General Public License as published by
 *                 the Free Software Foundation; either version 2 of the License, or
 *                 (at your option) any later version.
 *
 *                 jUPNP is distributed in the hope that it will be useful,
 *                 but WITHOUT ANY WARRANTY; without even the implied warranty of
 *                 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *                 GNU General Public License for more details.
 *
 *                 You should have received a copy of the GNU General Public License
 *                 along with jUPNP; if not, write to the Free Software
 *                 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * </pre>
 *
 *
 * This class stores our main config.
 *
 * This is a class that stores runtime configuration and common constants.
 *
 *
 * @author tex
 *
 * "Guns don't kill people. Chuck Norris kills People."
 * @link http://www.chucknorrisfacts.com/
 *
 */
public class Config {

	static Logger					logger						= Logger.getLogger(Config.class);

	public static final boolean		SILENT						= true;

	public static final String		CRLF						= "\r\n";

	private static final String		PRODUCT						= "jUPnP";

	private static final String		VERSION						= "0.1a";

	public static final String		SIGNATURE					= PRODUCT + "/" + VERSION;

	/**
	 * The Multicast Group as defined by UPnP
	 */
	public static final String		MULTICAST_GROUP				= "239.255.255.250";

	/**
	 * The Multicast Port as defined by UPnP
	 */
	public static final int			MULTICAST_PORT				= 1900;

	/**
	 * The first port we try to bind our webserver to. We start at 1025 because all ports below are not usable on UNIX
	 * systems.
	 */
	public static final int			HTTP_PORT_START				= 1025;

	private ServerSocket			httpServerSocket			= null;

	private int						httpServerPort				= 0;

	private InetAddress				localAddress				= null;

	private MulticastSocket			mcSocket					= null;

	private InetAddress				group						= null;

	/**
	 * This is random UUID used for our Device.
	 */
	private UUID					uuid						= UUID
																		.fromString("e22a72a3-2735-4feb-80af-131efad7cf4f");

	/**
	 * This will later hold our Device. This is only set when running in Device mode.
	 */
	private Device					device						= null;

	private DatagramSocket			httpUdpServerSocket			= null;

	/**
	 * This is used for the Cache-Control Header
	 */
	private int						cacheControlMaxAge			= 1800;

	/**
	 * An identification string for the switch power service
	 */
	public static final String		SWITCH_POWER_SERVICE_TYPE	= "urn:schemas-upnp-org:service:SwitchPower:1";

	/**
	 * An identification string for the switch power service
	 */
	public static final String		SWITCH_POWER_SERVICE_ID		= "urn:upnp-org:serviceId:SwitchPower.0001";

	/**
	 * An identification string for the dimming service
	 */
	public static final String		DIMMING_SERVICE_TYPE		= "urn:schemas-upnp-org:service:DimmingService:1";

	/**
	 * An identification string for the dimming service
	 */
	public static final String		DIMMING_SERVICE_ID			= "urn:upnp-org:serviceId:DimmingService.0001";

	/**
	 * Wait a random amout before replying to any request?
	 */
	private static final boolean	WAIT_RANDOM					= false;

	/**
	 * A unified random number generator.
	 */
	public Random					random						= new Random();

	/**
	 * @return the device
	 */
	public Device getDevice() {
		return device;
	}

	/**
	 *
	 * @return a String containig a proper Server: Header for HTTP-Responses
	 */
	public static String getHeaderServer() {
		return "SERVER: " + System.getProperty("os.name") + "/" + System.getProperty("os.version") + ", UPnP/1.0, "
				+ Config.SIGNATURE;
	}

	/**
	 * @param device
	 *            the device to set
	 */
	public void setDevice(Device device) {
		this.device = device;
	}

	/**
	 * @return the group
	 */
	public InetAddress getGroup() {
		return group;
	}

	/**
	 * The constructor initalizes the Multicast Socket so we don't have to do this everytime we use it.
	 *
	 */
	public Config() {
		try {
			this.localAddress = InetAddress.getLocalHost();
			this.mcSocket = new MulticastSocket(MULTICAST_PORT);
			group = InetAddress.getByName(MULTICAST_GROUP);
			this.mcSocket.joinGroup(group);
		} catch (UnknownHostException e) {
			logger.error(e);
		} catch (IOException e) {
			logger.error(e);
		}
	}

	/**
	 * This SHOULD be called when leaving the progam. Every Port will be closed.
	 *
	 */
	public void close() {
		try {
			this.mcSocket.leaveGroup(group);
			this.mcSocket.close();
			this.httpServerSocket.close();
			this.httpUdpServerSocket.close();
		} catch (IOException e) {
			logger.error(e);
		}
	}

	/**
	 * @return the localAddress
	 */
	public InetAddress getLocalAddress() {
		if (this.localAddress == null) {
			try {
				this.localAddress = InetAddress.getLocalHost();
			} catch (UnknownHostException e) {
				logger.trace("got unknown Host.Exception", e);
			}
		}
		return localAddress;
	}

	/**
	 * reads the stored config-file and sets the values
	 *
	 * @param configFile
	 * @return true if succesfull
	 * @throws IOException
	 */
	public boolean readConfig(String configFile) throws IOException {
		BufferedReader in = new BufferedReader(new FileReader(configFile));
		String currentLine = "";
		while ((currentLine = in.readLine()) != null) {
			if (!currentLine.startsWith("#")) { // skip comments
				StringTokenizer tokens = new StringTokenizer(currentLine);
				String currentToken = "";
				if (tokens.hasMoreTokens())
					currentToken = tokens.nextToken();
				if (currentToken.equals("Address")) {
					// this.setAddress(tokens.nextToken());
				} else if (currentToken.equals("Port")) {
					// this.setPort(Integer.parseInt(tokens.nextToken()));
				} else if (currentToken.equals("Username")) {
					// this.setUsername(tokens.nextToken());
				} else if (currentToken.equals("CacheMails") && tokens.nextToken().trim().equals("No".toLowerCase())) {
					// this.setCacheMails(false);
				} else if (currentToken.equals("Password")) {
					// this.setPassword(tokens.nextToken());
				} else if (currentToken.equals("ConnectRetries")) {
					// this.setConnectRetries(Integer.parseInt(tokens.nextToken()));
				}
			}
		}
		return true;
	}

	/**
	 * @return the mcSocket TODO Ist es ok wenn der MCSocket die ganz Zeit �ber offen ist?
	 */
	public synchronized MulticastSocket getMcSocket() {
		return mcSocket;
	}

	/**
	 * @return the uuid
	 */
	public UUID getUuid() {
		return uuid;
	}

	/**
	 * This method inintalizes the TCP and UDP ServerSockets and sets the http Port to a free one.
	 *
	 * @return the httpServerSocket
	 */
	public synchronized ServerSocket getHttpServerSocket() {
		if (this.httpServerSocket == null) {
			for (int port = HTTP_PORT_START; port < 65534; port++) {
				if (port == 65535) {
					throw new IllegalStateException();
				}
				try {
					this.httpServerSocket = new ServerSocket(port);
					this.httpUdpServerSocket = new DatagramSocket(port);
					this.httpServerPort = port;
					return this.httpServerSocket;
				} catch (IOException e) {
					// We ignore this exception and try the next port.
					this.httpServerPort = 0;
					this.httpUdpServerSocket = null;
					this.httpServerSocket = null;
				}
			}
		}
		logger.trace("getHttpServerSocket() - found free Port on " + this.httpServerPort);
		return this.httpServerSocket;
	}

	/**
	 * @return the httpUdpServerSocket
	 */
	public DatagramSocket getHttpUdpServerSocket() {
		if (this.httpUdpServerSocket == null) {
			this.getHttpServerSocket();
		}
		return httpUdpServerSocket;
	}

	/**
	 * @return the httpServerPort
	 */
	public int getHttpServerPort() {
		if (httpServerPort == 0) {
			this.getHttpServerSocket();
		}
		return httpServerPort;
	}

	/**
	 * @return the cacheControlMaxAge
	 */
	public int getCacheControlMaxAge() {
		return cacheControlMaxAge;
	}

	/**
	 * @param cacheControlMaxAge
	 *            the cacheControlMaxAge to set
	 */
	public void setCacheControlMaxAge(int cacheControlMaxAge) {
		this.cacheControlMaxAge = cacheControlMaxAge;
	}

	public static void waitRandom(Config config, int min, int max) {
		Config.waitRandom(config, min, max, false);
	}

	/**
	 * Wait an random amout of milicseconds.
	 * @param config Our Config
	 * @param min Wait at least min ms
	 * @param max Wat no more the max ms
	 * @param override Ignore Config.WAIT_RANDOM?
	 */
	public static void waitRandom(Config config, int min, int max, boolean override) {
		if (Config.WAIT_RANDOM || override) {
			long sleep = 0;
			if (min > max) {
				throw new IllegalArgumentException();
			}
			if (min != max) {
				sleep = min + config.random.nextInt(max - min);
			}
			try {
				Thread.sleep(sleep);
			} catch (InterruptedException e) {
				logger.trace("Got InterruptedExcpetion", e);
			}
			return;
		} else {
			return;
		}
	}
}
