package common.structures;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Namespace;

import common.core.Config;
import common.utils.XMLTools;

import device.multicast.NotifyChangeThread;
import device.multicast.SSDPAliveThread;
import device.multicast.SSDPByeByeThread;

/**
 * <pre>
 *             Copyright (c) 2007 Dominik Schulz
 *             Copyright (c) 2007 Florian Lindner
 *             Copyright (c) 2007 Philip Hartmann
 *
 *             This file is part of jUPnP.
 *
 *             jUPnP is free software; you can redistribute it and/or modify
 *             it under the terms of the GNU General Public License as published by
 *             the Free Software Foundation; either version 2 of the License, or
 *             (at your option) any later version.
 *
 *             jUPnP is distributed in the hope that it will be useful,
 *             but WITHOUT ANY WARRANTY; without even the implied warranty of
 *             MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *             GNU General Public License for more details.
 *
 *             You should have received a copy of the GNU General Public License
 *             along with jUPnP; if not, write to the Free Software
 *             Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * </pre>
 *
 * @author tex
 *
 * "Chuck Norris is currently suing NBC, claiming Law and Order are trademarked names for his left and right legs."
 */
public class Service {

	static Logger						logger			= Logger.getLogger(Service.class);

	private ArrayList<Action>			actions			= new ArrayList<Action>();

	private Config						config			= null;

	private String						controlURL		= "";

	private String						descriptionURL	= "";								// the only xml url

	private String						eventSubURL		= "";

	private String						serviceID		= "";

	private String						serviceType		= "";

	private int							specMinor		= 0;

	private int							specMajor		= 0;

	/**
	 * Contains all stateVariables of a Service
	 */
	private ArrayList<StateVariable>	stateVariables	= new ArrayList<StateVariable>();

	private Document					xmlSource		= null;

	/**
	 * Contains all subscriptions of controlpoints
	 */
	private ArrayList<Subscription>		subscriptions	= new ArrayList<Subscription>();

	/**
	 * The state variable shows wether the intel light is turned on
	 */
	public boolean setStateVariable(String varName, String value) {
		for (int i = 0; i < this.stateVariables.size(); i++) {
			if (this.stateVariables.get(i).getName().equalsIgnoreCase(varName)) {
				logger.trace("Set value of " + this.stateVariables.get(i).getName() + " to: " + value);
				this.stateVariables.get(i).setValue(value);
				this.notifyChange(varName, this.stateVariables.get(i).getValue());
				return true;
			}
		}
		return false;
	}

	public String getStateVariableValue(String name) {
		for (int i = 0; i < this.stateVariables.size(); i++) {
			if (this.stateVariables.get(i).getName().equalsIgnoreCase(name)) {
				return this.stateVariables.get(i).getValue();
			}
		}
		return "";
	}

	/**
	 * informs all subscribers about all state vars of this service
	 */
	public void notifyChangeAll() {
		this.cleanUp();
		for (int i = 0; i < this.stateVariables.size(); i++) {
			this.notifyChange(this.stateVariables.get(i).getName(), this.stateVariables.get(i).getValue());
		}
	}

	/**
	 * When there are internal changes the devices notifys its subscribers
	 * @param varName
	 * @param value
	 */
	private void notifyChange(String varName, String value) {
		this.cleanUp();
		// send a NOTIFY Request to each callback-url within this.subscriptions
		logger.trace("Sending NOTIFY to " + this.subscriptions.size() + " recipients.");
		for (int i = 0; i < this.subscriptions.size(); i++) {
			int seq = this.subscriptions.get(i).getSeq();
			NotifyChangeThread nct = new NotifyChangeThread(this.config, this.subscriptions.get(i).getCallbackUrl(),
					this.getURN() + "#" + i, varName, value, seq);
			Thread th = new Thread(nct);
			logger.trace("Sending NOTIFY to #" + i + ", SEQ: " + seq);
			th.start();
		}
	}

	/**
	 * creates a new and unique sid
	 *
	 * @return
	 */
	private String getNewSID() {
		String sid = this.getServiceType() + "#" + this.subscriptions.size();
		return sid;
	}

	/**
	 * makes the Controlpoint subscribe top the service of the device by sending a subscriptionrequest to the callback
	 * address
	 *
	 * @param callbackUrl
	 * @param timeOut
	 *            The Timeout in seconds!
	 * @return
	 */
	public String subscribe(String callbackUrl, String timeOut) {
		String sid = this.getNewSID();
		long to = 0;
		if (timeOut.startsWith("Second-")) {
			timeOut = timeOut.substring(7);
			to = Integer.parseInt(timeOut);
		}
		to = to * 1000; // convert to miliseconds
		Subscription subs = new Subscription(callbackUrl, sid, to);
		this.subscriptions.add(subs);
		logger.trace("Subscribed " + subs);
		this.notifyChangeAll();
		this.cleanUp();
		return sid;
	}

	/**
	 * Update the status of a subscription
	 *
	 * @param sid
	 * @return
	 */
	public boolean touch(String sid) {
		for (int i = 0; i < this.subscriptions.size(); i++) {
			if (this.subscriptions.get(i).getSid().equalsIgnoreCase(sid)) {
				logger.trace("touch( " + sid + ") - done!");
				this.subscriptions.get(i).touch();
				this.cleanUp();
				return true;
			}
		}
		this.cleanUp();
		logger.error("touch( " + sid + ") - NOT FOUND!");
		return false;
	}

	/**
	 * Update the timeout of a subscription
	 *
	 * @param sid
	 * @param timeOut
	 * @return
	 */
	public boolean updateTimeOut(String sid, String timeOut) {
		long to = 0;
		if (timeOut.startsWith("Second-")) {
			timeOut = timeOut.substring(7);
			to = Integer.parseInt(timeOut);
		}
		to = to * 1000; // convert to miliseconds
		for (int i = 0; i < this.subscriptions.size(); i++) {
			if (this.subscriptions.get(i).getSid().equalsIgnoreCase(sid)) {
				logger.trace("updateTimeOut( " + sid + ") - done!");
				this.subscriptions.get(i).setTimeOut(to);
				this.subscriptions.get(i).touch();
				this.cleanUp();
				return true;
			}
		}
		this.cleanUp();
		logger.error("updateTimeOut( " + sid + ") - NOT FOUND!");
		return false;
	}

	/**
	 * Remove all expired subscriptions.
	 *
	 */
	public void cleanUp() {
		logger.trace("Cleanup Running ...");
		for (int i = 0; i < this.subscriptions.size(); i++) {
			if (this.subscriptions.get(i).isExpired()) {
				logger.trace("Cleaning: " + this.subscriptions.get(i));
				this.unsubscribe(this.subscriptions.get(i).getSid());
			}
		}
	}

	/**
	 * a controlpoint can invoke unsubscribe to the service here
	 * @param sid
	 * @return
	 */
	public boolean unsubscribe(String sid) {
		for (int i = 0; i < this.subscriptions.size(); i++) {
			if (this.subscriptions.get(i).getSid().equalsIgnoreCase(sid)) {
				logger.trace("unsubscribe( " + sid + ") - removed!");
				return this.subscriptions.remove(i).getCallbackUrl().length() > 0;
			}
		}
		logger.trace("unsubscribe(" + sid + ") - NOT FOUND!");
		this.cleanUp();
		return false;
	}

	public Service(Config conf) {
		this.config = conf;
	}

	public boolean addAction(Action action) {
		return this.actions.add(action);
	}

	public void addStateVariable(StateVariable stv) {
		this.stateVariables.add(stv);
	}

	/**
	 * @return the actions
	 */
	public ArrayList<Action> getActions() {
		return actions;
	}

	/**
	 * @return the controlURL
	 */
	public String getControlURL() {
		if (controlURL.length() < 1) {
			this.controlURL = "_" + this.serviceType + "." + this.serviceID + "_control";
		}
		return controlURL;
	}

	/**
	 * @return the descriptionURL
	 */
	public String getDescriptionURL() {
		if (descriptionURL.length() < 1) {
			this.descriptionURL = "_" + this.serviceType + "." + this.serviceID + "_scpd.xml";
		}
		return descriptionURL;
	}

	/**
	 * @return the eventSubURL
	 */
	public String getEventSubURL() {
		if (eventSubURL.length() < 1) {
			this.eventSubURL = "_" + this.serviceType + "." + this.serviceID + "_event";
		}
		return eventSubURL;
	}

	public String getRealServiceID() {
		return this.serviceID;
	}

	public String getRealServiceType() {
		return this.serviceType;
	}

	/**
	 * @return the serviceID
	 */
	public String getServiceID() {
		return serviceID;
	}

	/**
	 * @return the serviceType
	 */
	public String getServiceType() {
		return serviceType;
	}

	public int getSpecMajor() {
		return specMajor;
	}

	public int getSpecMinor() {
		return specMinor;
	}

	/**
	 * @return the stateVariables
	 */
	public ArrayList<StateVariable> getStateVariables() {
		return stateVariables;
	}

	public String getURN() {
		return this.getServiceType();
	}

	public String getUUID() {
		return "uuid:" + this.config.getUuid();
	}

	/**
	 * @return a XML-Structure (Element) representing this node
	 */
	public Element getXMLElements() {
		Element root = new Element("service");

		Element serviceType = new Element("serviceType");
		serviceType.setText(this.getServiceType());
		root.addContent(serviceType);

		Element serviceID = new Element("serviceID");
		serviceID.setText(this.getServiceID());
		root.addContent(serviceID);

		Element controlURL = new Element("controlURL");
		controlURL.setText(this.getControlURL().toString());
		root.addContent(controlURL);

		Element eventSubURL = new Element("eventSubURL");
		eventSubURL.setText(this.getEventSubURL().toString());
		root.addContent(eventSubURL);

		Element descriptionURL = new Element("SCPDURL");
		descriptionURL.setText(this.getDescriptionURL().toString());
		root.addContent(descriptionURL);

		Element actions = new Element("actionList");
		for (int i = 0, n = this.actions.size(); i < n; i++) {
			actions.addContent(this.actions.get(i).getXMLElements());
		}
		root.addContent(actions);

		Element stateVariables = new Element("serviceStateTable");
		for (int i = 0, n = this.stateVariables.size(); i < n; i++) {
			stateVariables.addContent(this.stateVariables.get(i).getXMLElements());
		}
		root.addContent(stateVariables);
		return root;
	}

	/**
	 * Form a XML-description
	 * @return
	 */
	public Element getXMLElementsForDevice() {
		Element root = new Element("service");

		Element serviceType = new Element("serviceType");
		serviceType.setText(this.getServiceType());
		root.addContent(serviceType);

		Element serviceID = new Element("serviceId");
		serviceID.setText(this.getServiceID());
		root.addContent(serviceID);

		Element controlURL = new Element("controlURL");
		controlURL.setText(this.getControlURL().toString());
		root.addContent(controlURL);

		Element eventSubURL = new Element("eventSubURL");
		eventSubURL.setText(this.getEventSubURL().toString());
		root.addContent(eventSubURL);

		Element descriptionURL = new Element("SCPDURL");
		descriptionURL.setText(this.getDescriptionURL().toString());
		root.addContent(descriptionURL);

		return root;
	}

	/**
	 * Creates the SOAP Envelope
	 * @return SOAP XML Envelope
	 */
	public Element getXMLElementsWithEnvelope() {
		Element root = new Element("scpd");
		Namespace ns = Namespace.getNamespace("urn:schemas-upnp-org:service-1-0");
		root.setNamespace(ns);
		Element specVersion = new Element("specVersion");
		Element major = new Element("major");
		major.setText("1");
		Element minor = new Element("minor");
		minor.setText("0");
		specVersion.addContent(major);
		specVersion.addContent(minor);
		root.addContent(specVersion);
		Element actions = new Element("actionList");
		for (int i = 0, n = this.actions.size(); i < n; i++) {
			actions.addContent(this.actions.get(i).getXMLElements());
		}
		root.addContent(actions);

		Element stateVariables = new Element("serviceStateTable");
		for (int i = 0, n = this.stateVariables.size(); i < n; i++) {
			stateVariables.addContent(this.stateVariables.get(i).getXMLElements());
		}
		root.addContent(stateVariables);
		return root;
	}

	public Document getXmlSource() {
		return xmlSource;
	}

	/**
	 *
	 * @return a pretty formated XML String representing this node
	 */
	public String getXMLString() {
		return XMLTools.xmlToString(this.getXMLElements());
	}

	/**
	 * send an alive message (multicast SSDP)
	 *
	 */
	public void sendSSDPAlive() {
		SSDPAliveThread ssdpat = new SSDPAliveThread(this.config, this.getURN(), this.getUUID() + "::" + this.getURN());
		Thread t = new Thread(ssdpat);
		t.start();
	}

	/**
	 * send an "byebye" message (multicast SSDP)
	 *
	 */
	public void sendSSDPByeBye() {
		SSDPByeByeThread ssdpbbt = new SSDPByeByeThread(this.config, this.getURN(), this.getUUID() + "::"
				+ this.getURN());
		Thread t = new Thread(ssdpbbt);
		t.start();

	}

	/**
	 * @param actions
	 *            the actions to set
	 */
	public void setActions(ArrayList<Action> actions) {
		this.actions = actions;
	}

	/**
	 * @param controlURL
	 *            the controlURL to set
	 */
	public void setControlURL(String controlURL) {
		this.controlURL = controlURL;
	}

	/**
	 * @param descriptionURL
	 *            the descriptionURL to set
	 */
	public void setDescriptionURL(String descriptionURL) {
		this.descriptionURL = descriptionURL;
	}

	/**
	 * @param eventSubURL
	 *            the eventSubURL to set
	 */
	public void setEventSubURL(String eventSubURL) {
		this.eventSubURL = eventSubURL;
	}

	/**
	 * @param serviceID
	 *            the serviceID to set
	 */
	public void setServiceID(String serviceID) {
		this.serviceID = serviceID;
	}

	/**
	 * @param serviceType
	 *            the serviceType to set
	 */
	public void setServiceType(String serviceType) {
		this.serviceType = serviceType;
	}

	public void setSpecMajor(int specMajor) {
		this.specMajor = specMajor;
	}

	public void setSpecMinor(int specMinor) {
		this.specMinor = specMinor;
	}

	/**
	 * @param stateVariables
	 *            the stateVariables to set
	 */
	public void setStateVariables(ArrayList<StateVariable> stateVariables) {
		this.stateVariables = stateVariables;
	}

	/**
	 * setting up the XML sources, by parsing the XML file
	 * @param xmlSource
	 */
	public void setXmlSource(Document xmlSource) {
		this.xmlSource = xmlSource;
		List mainNodes = this.getXmlSource().getRootElement().getChildren();
		// traverse the XML-tree
		for (Iterator iter = mainNodes.iterator(); iter.hasNext();) {
			Element mainElement = (Element) iter.next();
			if (mainElement == null) {
				continue;
			}
			logger.trace(mainElement.toString());
			if (mainElement.getName().equalsIgnoreCase("specVersion")) {
				List specNodes = mainElement.getChildren();
				for (Iterator specIter = specNodes.iterator(); specIter.hasNext();) {
					Element specEntry = (Element) specIter.next();
					if (specEntry == null) {
						continue;
					}
					if (specEntry.getName().equals("major")) {
						this.setSpecMajor(Integer.parseInt(specEntry.getText()));
					}
					if (specEntry.getName().equals("minor")) {
						this.setSpecMinor(Integer.parseInt(specEntry.getText()));
					}
				}
			}
			// get to the action lists
			if (mainElement.getName().equalsIgnoreCase("actionList")) {
				List actionNodes = mainElement.getChildren();
				for (Iterator actionIter = actionNodes.iterator(); actionIter.hasNext();) {
					Element actionNode = (Element) actionIter.next();
					if (actionNode == null) {
						continue;
					}
					logger.trace("Element 2:" + actionNode.toString());

					Action act = new Action();
					List actionEntrys = actionNode.getChildren();
					for (Iterator iter3 = actionEntrys.iterator(); iter3.hasNext();) {
						Element actionEntry = (Element) iter3.next();
						if (actionEntry == null) {
							continue;
						}
						if (actionEntry.getName().equalsIgnoreCase("name")) {
							act.setName(actionEntry.getText());
						}
						if (actionEntry.getName().equalsIgnoreCase("argumentList")) {
							List argumentList = actionEntry.getChildren();
							for (Iterator argumentIter = argumentList.iterator(); argumentIter.hasNext();) {
								Element argumentNode = (Element) argumentIter.next();
								if (argumentNode == null) {
									continue;
								}
								Argument arg = new Argument();
								List argumentEntries = argumentNode.getChildren();
								for (Iterator iter5 = argumentEntries.iterator(); iter5.hasNext();) {
									Element argumentEntry = (Element) iter5.next();
									if (argumentEntry == null) {
										continue;
									}
									if (argumentEntry.getName().equalsIgnoreCase("name")) {
										arg.setName(argumentEntry.getText());
									}
									if (argumentEntry.getName().equalsIgnoreCase("direction")) {
										arg.setDirection(argumentEntry.getText());
									}
									if (argumentEntry.getName().equalsIgnoreCase("relatedStateVariable")) {
										arg.setRstv(argumentEntry.getText());
									}
								}
								act.addArgument(arg);
							}
						}
					}
					this.addAction(act);
				}
			}
			if (mainElement.getName().equalsIgnoreCase("serviceStateTable")) {
				List sstEntryList = mainElement.getChildren();
				for (Iterator sstEntryIter = sstEntryList.iterator(); sstEntryIter.hasNext();) {
					Element sstEntry = (Element) sstEntryIter.next();
					if (sstEntry == null) {
						continue;
					}
					if (sstEntry.getName().equalsIgnoreCase("stateVariable")) {
						StateVariable stv = new StateVariable();

						String attrib = sstEntry.getAttributeValue("sendEvents");
						if (attrib != null) {
							stv.setSendingEvents(attrib.equalsIgnoreCase("yes"));
						}
						List svEntryList = sstEntry.getChildren();
						for (Iterator svEntryIter = svEntryList.iterator(); svEntryIter.hasNext();) {
							Element svEntry = (Element) svEntryIter.next();
							if (svEntry == null) {
								continue;
							}
							if (svEntry.getName().equalsIgnoreCase("name")) {
								stv.setName(svEntry.getText());
							}
							if (svEntry.getName().equalsIgnoreCase("dataType")) {
								stv.setDataType(svEntry.getText());
							}
							if (svEntry.getName().equalsIgnoreCase("allowedValueRange")) {
								List svEntrysList = svEntry.getChildren();
								for (Iterator svEntrysIter = svEntrysList.iterator(); svEntrysIter.hasNext();) {
									Element svEntryAllowedValue = (Element) svEntrysIter.next();
									if (svEntryAllowedValue == null) {
										continue;
									}
									AllowedValue av = new AllowedValue(svEntryAllowedValue.getName(),
											svEntryAllowedValue.getText());
									stv.getAllowedValues().add(av);
								}
							}
							if (svEntry.getName().equalsIgnoreCase("allowedValueList")) {
								List svEntrysList = svEntry.getChildren();
								for (Iterator svEntrysIter = svEntrysList.iterator(); svEntrysIter.hasNext();) {
									Element svEntryAllowedValue = (Element) svEntrysIter.next();
									if (svEntryAllowedValue == null) {
										continue;
									}
									if (svEntryAllowedValue.getName().equalsIgnoreCase("minimum")) {
										stv.setMinimum(Integer.parseInt(svEntryAllowedValue.getText()));
									}
									if (svEntryAllowedValue.getName().equalsIgnoreCase("maximum")) {
										stv.setMinimum(Integer.parseInt(svEntryAllowedValue.getText()));
									}
									AllowedValue av = new AllowedValue(svEntryAllowedValue.getName(),
											svEntryAllowedValue.getText());
									stv.getAllowedValues().add(av);
								}
							}
						}
						this.addStateVariable(stv);
					}
				}
			}
		}
	}

}
