package controlpoint.http;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.InetAddress;
import java.net.Socket;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;
import org.jdom.Document;
import org.jdom.input.SAXBuilder;

import common.core.Config;
import common.http.ConfigHttpd;
import common.http.Header;
import common.utils.XMLTools;

import controlpoint.core.ControlPoint;
import controlpoint.structures.RemoteDevice;

/**
 * <pre>
 *           Copyright (c) 2007 Dominik Schulz
 *           Copyright (c) 2007 Florian Lindner
 *           Copyright (c) 2007 Philip Hartmann
 *
 *           This file is part of jUPnP.
 *
 *           jUPnP is free software; you can redistribute it and/or modify
 *           it under the terms of the GNU General Public License as published by
 *           the Free Software Foundation; either version 2 of the License, or
 *           (at your option) any later version.
 *
 *           jUPnP is distributed in the hope that it will be useful,
 *           but WITHOUT ANY WARRANTY; without even the implied warranty of
 *           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *           GNU General Public License for more details.
 *
 *           You should have received a copy of the GNU General Public License
 *           along with jUPnP; if not, write to the Free Software
 *           Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * </pre>
 *
 * @author Dominik
 * @author Philip
 * @author Florian
 *
 * "Chuck Norris can lead a horse to water AND make it drink."
 */
public final class HttpRequestCP implements Runnable {

	static Logger	logger	= Logger.getLogger(HttpRequestCP.class);

	/**
	 * This method implements our own readline which we were advised to use instead of the readers readline. the
	 * difference is simple: readline looks for a New-Line (\n) character. We look for a combination of Carriage-Return
	 * and New-Line (\r\n) because a New-Line can occur within a line.
	 *
	 * @param br
	 *            the BufferedReader to read from
	 * @return the line which was read from the reader
	 * @throws IOException
	 */
	private static String crlfReadline(BufferedReader br) throws IOException {
		int buffer = -1;
		StringBuilder currentLine = new StringBuilder();
		while ((buffer = br.read()) != -1) {
			if (buffer != "\r".toCharArray()[0]) {
				currentLine.append((char) buffer);
			} else {
				if (br.read() == "\n".toCharArray()[0]) {
					return currentLine.toString();
				} else {
					currentLine.append("\n");
				}
			}
		}
		return currentLine.toString();
	}

	@SuppressWarnings("unused")
	private static void sendBytes(FileInputStream fis, OutputStream os) throws IOException {
		// Construct a 1K buffer to hold bytes on their way to the socket.
		byte[] buffer = new byte[1024];
		int bytes = 0;
		// Copy requested file into the socket's output stream.
		while ((bytes = fis.read(buffer)) != -1)
			os.write(buffer, 0, bytes);
	}

	private Config				config;

	private ConfigHttpd			configHttpd;

	/**
	 * The value of the Content-Length: header field
	 */
	private int					contentLength		= 0;

	/**
	 * The value of the Host: header field
	 */
	private String				host				= "";

	public static final String	locationStr			= "LOCATION:";

	private String				location			= "";

	public static final String	ntStr				= "NT:";

	private String				nt					= "";

	public static final String	ntsStr				= "NTS:";

	private String				nts					= "";

	public static final String	serverStr			= "SERVER:";

	private String				server				= "";

	public static final String	usnStr				= "USN:";

	private String				usn					= "";

	public static final String	cacheControlStr		= "CACHE-CONTROL:";

	private String				cacheControl		= "";

	public static final String	stStr				= "ST:";

	private String				st					= "";

	public static final String	mxStr				= "MX:";

	private String				mx					= "";

	public static final String	manStr				= "MAN:";

	private String				man					= "";

	public static final String	sidStr				= "SID:";

	private String				sid					= "";

	public static final String	subscriptionIdStr	= "Subscription-ID:";

	private String				subscriptionId		= "";

	public static final String	callBackStr			= "Call-Back:";

	private String				callBack			= "";

	public static final String	ScopeStr			= "Scope:";

	private String				scope				= "";

	public static final String	timeOutStr			= "Timeout:";

	private String				timeOut				= "";

	public static final String	optStr				= "OPT:";

	private String				opt					= "";

	public static final String	o1nlsStr			= "01-NLS_";

	private String				o1nls				= "";

	InputStream					is					= null;

	OutputStream				os					= null;

	byte[]						output				= null;

	private Socket				socket;

	private ControlPoint		cp					= null;

	@SuppressWarnings("unused")
	private InetAddress			senderAddress		= null;

	@SuppressWarnings("unused")
	private int					senderPort			= 0;

	public HttpRequestCP(byte[] input, byte[] output, ConfigHttpd configHttpd, Config conf) {
		this.configHttpd = configHttpd;
		this.config = conf;
		this.is = new ByteArrayInputStream(input);
		this.os = new ByteArrayOutputStream();
		this.output = output;
	}

	public HttpRequestCP(byte[] input, byte[] output, ConfigHttpd configHttpd, Config conf, ControlPoint cp) {
		this.configHttpd = configHttpd;
		this.config = conf;
		this.is = new ByteArrayInputStream(input);
		this.os = new ByteArrayOutputStream();
		this.output = output;
		this.cp = cp;
	}

	public HttpRequestCP(byte[] input, byte[] output, ConfigHttpd configHttpd, Config conf, ControlPoint cp,
			InetAddress senderAddress, int senderPort) {
		this.configHttpd = configHttpd;
		this.config = conf;
		this.is = new ByteArrayInputStream(input);
		this.os = new ByteArrayOutputStream();
		this.output = output;
		this.cp = cp;
		this.senderAddress = senderAddress;
		this.senderPort = senderPort;
	}

	public HttpRequestCP(Socket socket, ConfigHttpd configHttpd, Config conf, ControlPoint cp) throws Exception {
		this.socket = socket;
		this.configHttpd = configHttpd;
		this.config = conf;
		this.cp = cp;
	}

	@SuppressWarnings("unused")
	private String contentType(String fileName) {
		StringTokenizer tokens = new StringTokenizer(fileName, ".");
		String currentToken = "";
		while (tokens.hasMoreTokens()) {
			currentToken = tokens.nextToken(); // we are only interested in the
			// last token
		}
		return this.configHttpd.getMimeType(currentToken);
	}

	/**
	 * handling an http-request
	 * @throws Exception
	 */
	private void processRequest() throws Exception {
		InputStreamReader isr = null;
		BufferedReader br = null;

		OutputStreamWriter osw = null;
		BufferedWriter bw = null;

		try {
			configHttpd.countUp();

			if (is == null) {
				is = socket.getInputStream();
			}
			isr = new InputStreamReader(is);
			br = new BufferedReader(isr);
			if (os == null) {
				os = socket.getOutputStream();
				osw = new OutputStreamWriter(os);
				bw = new BufferedWriter(osw);
			}

			String requestLine = crlfReadline(br);
			// To see what's going on we print each HTTP Request (HTTP, HTTPU or HTTPMU) to the log.
			logger.trace("HTTP-REQUEST: " + requestLine);

			String headerLine = null; // read the request header
			while ((headerLine = crlfReadline(br)).length() != 0) {
				if (headerLine.startsWith("Content-Length:")) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					// String: Content-Length:
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					// the content-length
					if (tokens.hasMoreTokens())
						this.contentLength = Integer.parseInt(tokens.nextToken());
				} else if (headerLine.startsWith("Host:")) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.host = tokens.nextToken();
				} else if (headerLine.startsWith(locationStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.location = tokens.nextToken();
				} else if (headerLine.startsWith(ntStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.nt = tokens.nextToken();
				} else if (headerLine.startsWith(ntsStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.nts = tokens.nextToken();
				} else if (headerLine.startsWith(serverStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.server = tokens.nextToken();
				} else if (headerLine.startsWith(usnStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.usn = tokens.nextToken();
				} else if (headerLine.startsWith(cacheControlStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.cacheControl = tokens.nextToken();
				} else if (headerLine.startsWith(stStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.st = tokens.nextToken();
				} else if (headerLine.startsWith(mxStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.mx = tokens.nextToken();
				} else if (headerLine.startsWith(manStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.man = tokens.nextToken();
				} else if (headerLine.startsWith(sidStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.sid = tokens.nextToken();
				} else if (headerLine.startsWith(subscriptionIdStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.subscriptionId = tokens.nextToken();
				} else if (headerLine.startsWith(callBackStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.callBack = tokens.nextToken();
				} else if (headerLine.startsWith(ScopeStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.scope = tokens.nextToken();
				} else if (headerLine.startsWith(timeOutStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.timeOut = tokens.nextToken();
				} else if (headerLine.startsWith(optStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.opt = tokens.nextToken();
				} else if (headerLine.startsWith(o1nlsStr)) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.o1nls = tokens.nextToken();
					logger.trace("O1NLS: " + this.o1nls);
				}
			}

			String httpMethod = "";
			String fileName = "";
			String httpVersion = "";

			StringTokenizer tokens = new StringTokenizer(requestLine);
			if (tokens.hasMoreTokens()) {
				// GET, HEAD, POST, SUBSCRIBE, UNSUBSCRIBE or NOTIFY
				httpMethod = tokens.nextToken();
			}
			if (tokens.hasMoreTokens()) {
				fileName = tokens.nextToken();
			}
			if (tokens.hasMoreTokens()) {
				httpVersion = tokens.nextToken();
				if (!httpVersion.equalsIgnoreCase("HTTP/1.1")) {
					logger.trace("Wrong HTTP-Version (" + httpVersion + " instead of HTTP/1.1) - ignoring.");
				}
			}

			// build the header
			Header header = new Header(os);

			// Split filename
			StringTokenizer tokens1 = new StringTokenizer(fileName, "?");
			String realFileName = "";
			@SuppressWarnings("unused")
			String arguments = "";
			if (tokens1.hasMoreTokens()) {
				realFileName = tokens1.nextToken();
			}
			if (tokens1.hasMoreTokens()) {
				arguments = tokens1.nextToken();
			}

			String entityBody = Config.CRLF;

			if (httpMethod.equals(ConfigHttpd.HTTP_METHOD_GET)) {
				/*
				 * ################################################################
				 * This code handles the HTTPM/U Method GET
				 * ################################################################
				 */
				logger.trace("processRequest() - HTTP_GET");
				// send the root-xml-file
				if (realFileName.equals("/")) {
					header.addHeaderLine(ConfigHttpd.HTTP_STATUS_200);
					header.addHeaderLine("Content-type: text/xml" + "; " + configHttpd.getDefaultCharset());
					entityBody = XMLTools.xmlToString(this.config.getDevice().getXMLElements());
					header.sendHeader();
					os.write(entityBody.getBytes());
				} else if (realFileName.endsWith("_scpd.xml")) { // return SCPD-File
					logger.trace("SCPD: " + realFileName);
					String string = realFileName.substring(2, realFileName.length() - 9);
					String cleanFilename = realFileName.substring(1);
					logger.trace("SCPD-URL: " + string);
					StringTokenizer tokens2 = new StringTokenizer(string, ".");
					@SuppressWarnings("unused")
					String service = tokens2.nextToken();
					@SuppressWarnings("unused")
					String serviceId = tokens2.nextToken();
					logger.trace("SCPD-Service: " + service);
					logger.trace("SCPD-ServiceId: " + serviceId);
					header.addHeaderLine(ConfigHttpd.HTTP_STATUS_200);
					header.addHeaderLine("Content-type: text/xml" + "; " + configHttpd.getDefaultCharset());
					entityBody = XMLTools.xmlToString(this.config.getDevice().getServiceByURL(cleanFilename)
							.getXMLElementsWithEnvelope());
					header.sendHeader();
					os.write(entityBody.getBytes());
				} else if (realFileName.endsWith("_control")) { // return control file
					logger.trace("Control: " + realFileName);
				} else if (realFileName.endsWith("_event")) { // retrun event file
					logger.trace("Event: " + realFileName);
				} else if (false) {
					// placeholder for a different file
				} else {
					// return a error 404-not found
					// 404 we cant read
					header.addHeaderLine(ConfigHttpd.HTTP_STATUS_404);
					header.addHeaderLine("Content-type: text/html" + "; " + configHttpd.getDefaultCharset());
					entityBody = "<html><head><title>Not Found</title></head>" + "<body><h1>Error 404 - Not Found</h1>"
							+ ConfigHttpd.SIGNATURE + "</body></html>";
					header.sendHeader();
					os.write(entityBody.getBytes());
				}
			} else if (httpMethod.equals(ConfigHttpd.HTTPMU_METHOD_MSEARCH)) {
				/*
				 * ################################################################
				 * This code handles the HTTPM/U Method M-SEARCH
				 * ################################################################
				 */
				logger.trace("processRequest() - HTTP_M-SEARCH");
				header.addHeaderLine(ConfigHttpd.HTTP_STATUS_200);
				header.addHeaderLine("ST: upnp:rootdevice");
				header.addHeaderLine("EXT: ");
				header.addHeaderLine(Config.getHeaderServer());
				header.addHeaderLine("USN: uuid:" + config.getUuid() + "::upnp:rootdevice");
				header.addHeaderLine("CACHE-CONTROL: max-age=3600");
				header.addHeaderLine("LOCATION: http://" + this.config.getLocalAddress().getHostAddress() + ":"
						+ this.config.getHttpServerPort() + "/");
				// wait between 300 and 1000 ms
				Config.waitRandom(this.config, 30, 100);
				header.sendHeader();
				logger.trace("Sent Header in reply to M-SEARCH: " + header.toString());
				os.write(entityBody.getBytes());
				this.output = ((ByteArrayOutputStream) this.os).toByteArray();
			} else if (httpMethod.equals(ConfigHttpd.HTTPMU_METHOD_NOTIFY)) {
				/*
				 * ################################################################
				 * This code handles the HTTPM/U Method NOTIFY
				 * ################################################################
				 */
				logger.trace("processRequest() - HTTP_NOTIFY - NTS: " + nts + ", NT: " + nt);
				logger.trace("processRequest() - Location: " + location);
				String reqEntityBody = "";
				int buffer = -1;
				while ((reqEntityBody.length() < this.contentLength) && ((buffer = br.read()) != -1)) {
					reqEntityBody += (char) buffer;
				}

				logger.trace("processRequest() - NOTIFY - XML: " + reqEntityBody);
				if (reqEntityBody.length() > 0) {
					Document xml = null;
					SAXBuilder xmlBuilder = new SAXBuilder();
					ByteArrayInputStream bais = new ByteArrayInputStream(reqEntityBody.getBytes());
					xml = xmlBuilder.build(bais);
					logger.trace(XMLTools.xmlToString(xml));
					int loadLevel = 0;
					int powerStatus = 0;
					try {
						String tmp = "";
						tmp = XMLTools.getXMLNode(xml.getRootElement().getChildren(), "LoadLevelStatus").getText();
						loadLevel = Integer.parseInt(tmp);
					} catch (NullPointerException e) {
						// just try to extract the info
						loadLevel = this.cp.getLoadLevel();
						// if the parsing fails, we set the loadLevel to its previous value
					}
					try {
						String tmp = "";
						tmp = XMLTools.getXMLNode(xml.getRootElement().getChildren(), "Status").getText();
						powerStatus = Integer.parseInt(tmp);
					} catch (NullPointerException e) {
						// just try to extract the info
						powerStatus = this.cp.getPowerStatus();
						// if the parsing fails, we set the loadLevel to its previous value
					}
					logger.trace("LoadLevel: " + loadLevel);
					logger.trace("PowerStatus: " + powerStatus);
					this.cp.setLoadLevel(loadLevel);
					this.cp.setPowerStatus(powerStatus);
				}
				header.addHeaderLine(ConfigHttpd.HTTP_STATUS_200);

				// Step 2: we received the NOTIFY-msg and created the RemoteDevice
				RemoteDevice rdev = new RemoteDevice(false, config);
				rdev.setCacheControl(cacheControl);
				rdev.setCallBack(callBack);
				rdev.setContentLength(contentLength);
				rdev.setHost(host);
				rdev.setLocation(location);
				rdev.setMan(man);
				rdev.setMx(mx);
				rdev.setNt(nt);
				rdev.setNts(nts);
				rdev.setOpt(opt);
				rdev.setScope(scope);
				rdev.setServer(server);
				rdev.setSid(sid);
				rdev.setSt(st);
				rdev.setSubscriptionId(subscriptionId);
				rdev.setTimeOut(timeOut);
				rdev.setUsn(usn);

				// add the device to the ControlPoint
				// Step 3: requesting the XML done in handleRemoteDevice
				if (this.cp == null) {
					// logger.debug("Not running as control-point!");
				} else {
					this.cp.handleRemoteDevice(rdev);
				}
				// wait between 30 and 100 ms
				Config.waitRandom(this.config, 30, 100);
				header.sendHeader();
				os.write(entityBody.getBytes());
			} else if (httpMethod.equals(ConfigHttpd.HTTPMU_METHOD_SUBSCRIBE)) {
				/*
				 * ################################################################
				 * This code handles the HTTPM/U Method SUBSCRIBE
				 * ################################################################
				 */
				logger.trace("processRequest() - HTTP_SUBSCRIBE");
				header.addHeaderLine(ConfigHttpd.HTTP_STATUS_400);
				header.sendHeader();
				os.write(entityBody.getBytes());
			} else if (httpMethod.equals(ConfigHttpd.HTTPMU_METHOD_UNSUBSCRIBE)) {
				/*
				 * ################################################################
				 * This code handles the HTTPM/U Method UNSUBSCRIBE
				 * ################################################################
				 */
				logger.trace("processRequest() - HTTP_UNSUBSCRIBE");
				header.addHeaderLine(ConfigHttpd.HTTP_STATUS_400);
				header.sendHeader();
				os.write(entityBody.getBytes());
			} else {
				/*
				 * ################################################################
				 * This code handles every other
				 * HTTPM/U Method (by returning a Code 500 - Internal Server Error)
				 * ################################################################
				 */
				header.addHeaderLine(ConfigHttpd.HTTP_STATUS_500);
				header.sendHeader();
				os.write(entityBody.getBytes());
			}
		} catch (Exception e) {
			throw e;
		} finally {
			try {
				bw.close();
			} catch (Exception e) {
				logger.trace(e);
			}

			try {
				os.close();
			} catch (Exception e) {
				logger.trace(e);
			}

			try {
				br.close();
			} catch (Exception e) {
				logger.trace(e);
			}

			try {
				socket.close();
			} catch (Exception e) {
				logger.trace(e);
			}
			try {
				this.output = ((ByteArrayOutputStream) this.os).toByteArray();
			} catch (Exception e) {
				logger.trace(e);
			}
			configHttpd.countDown();
		}
	}

	/**
	 * @see common.http.IHttpRequest#run()
	 */
	public void run() {
		try {
			processRequest();
		} catch (Exception e) {
			logger.trace(e);
		}
	}

	/**
	 * @return the output
	 */
	public byte[] getOutput() {
		return output;
	}

}
