package device.http;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.InetAddress;
import java.net.Socket;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;
import org.jdom.Document;
import org.jdom.input.SAXBuilder;

import common.core.Config;
import common.http.ConfigHttpd;
import common.http.Header;
import common.structures.Service;
import common.utils.XMLTools;

import controlpoint.core.ControlPoint;
import device.multicast.DiscoverReplyThread;

/**
 * <pre>
 *                         Copyright (c) 2007 Dominik Schulz
 *                         Copyright (c) 2007 Florian Lindner
 *                         Copyright (c) 2007 Philip Hartmann
 *
 *                         This file is part of jHTTPd.
 *
 *                         jHTTPd is free software; you can redistribute it and/or modify
 *                         it under the terms of the GNU General Public License as published by
 *                         the Free Software Foundation; either version 2 of the License, or
 *                         (at your option) any later version.
 *
 *                         jHTTPd is distributed in the hope that it will be useful,
 *                         but WITHOUT ANY WARRANTY; without even the implied warranty of
 *                         MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *                         GNU General Public License for more details.
 *
 *                         You should have received a copy of the GNU General Public License
 *                         along with jHTTPd; if not, write to the Free Software
 *                         Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * </pre>
 *
 * This class represents a single HttpRequest (Thread). It gets created when the serversocket receives a new connection
 * and then this is handled.
 *
 * @author Dominik
 * @author Philip
 * @author Florian
 *
 * "Superman once watched an episode of Walker, Texas Ranger. He then cried himself to sleep."
 */
public final class HttpRequestDev implements Runnable {

	static Logger				logger				= Logger.getLogger(HttpRequestDev.class);

	public static final String	locationStr			= "LOCATION:";

	public static final String	ntStr				= "NT:";

	public static final String	ntsStr				= "NTS:";

	public static final String	serverStr			= "SERVER:";

	public static final String	usnStr				= "USN:";

	public static final String	cacheControlStr		= "CACHE-CONTROL:";

	public static final String	stStr				= "ST:";

	public static final String	mxStr				= "MX:";

	public static final String	manStr				= "MAN:";

	public static final String	sidStr				= "SID:";

	public static final String	subscriptionIdStr	= "Subscription-ID:";

	public static final String	callBackStr			= "Callback:";

	public static final String	ScopeStr			= "Scope:";

	public static final String	timeOutStr			= "Timeout:";

	public static final String	optStr				= "OPT:";

	public static final String	o1nlsStr			= "01-NLS_";

	/**
	 * This method implements our own readline which we were advised to use instead of the readers readline. the
	 * difference is simple: readline looks for a New-Line (\n) character. We look for a combination of Carriage-Return
	 * and New-Line (\r\n) because a New-Line can occur within a line.
	 *
	 * @param br
	 *            the BufferedReader to read from
	 * @return the line which was read from the reader
	 * @throws IOException
	 */
	private static String crlfReadline(BufferedReader br) throws IOException {
		int buffer = -1;
		StringBuilder currentLine = new StringBuilder();
		while ((buffer = br.read()) != -1) {
			if (buffer != "\r".toCharArray()[0]) {
				currentLine.append((char) buffer);
			} else {
				if (br.read() == "\n".toCharArray()[0]) {
					return currentLine.toString();
				} else {
					currentLine.append("\n");
				}
			}
		}
		return currentLine.toString();
	}

	@SuppressWarnings("unused")
	private static void sendBytes(FileInputStream fis, OutputStream os) throws IOException {
		// Construct a 1K buffer to hold bytes on their way to the socket.
		byte[] buffer = new byte[1024];
		int bytes = 0;
		// Copy requested file into the socket's output stream.

		while ((bytes = fis.read(buffer)) != -1)
			os.write(buffer, 0, bytes);
	}

	private Config			config;

	private ConfigHttpd		configHttpd;

	/**
	 * The value of the Content-Length: header field
	 */
	private int				contentLength	= 0;

	/**
	 * The value of the Host: header field
	 */
	@SuppressWarnings("unused")
	private String			host			= "";

	@SuppressWarnings("unused")
	private String			location		= "";

	@SuppressWarnings("unused")
	private String			nt				= "";

	@SuppressWarnings("unused")
	private String			nts				= "";

	@SuppressWarnings("unused")
	private String			server			= "";

	@SuppressWarnings("unused")
	private String			usn				= "";

	@SuppressWarnings("unused")
	private String			cacheControl	= "";

	@SuppressWarnings("unused")
	private String			st				= "";

	@SuppressWarnings("unused")
	private String			mx				= "";

	@SuppressWarnings("unused")
	private String			man				= "";

	private String			sid				= "";

	@SuppressWarnings("unused")
	private String			subscriptionId	= "";

	private String			callBack		= "";

	@SuppressWarnings("unused")
	private String			scope			= "";

	@SuppressWarnings("unused")
	private String			timeOut			= "";

	@SuppressWarnings("unused")
	private String			opt				= "";

	@SuppressWarnings("unused")
	private String			o1nls			= "";

	InputStream				is				= null;

	OutputStream			os				= null;

	byte[]					output			= null;

	private Socket			socket;

	@SuppressWarnings("unused")
	private ControlPoint	cp				= null;

	private InetAddress		senderAddress	= null;

	private int				senderPort		= 0;

	private HttpRequestDev() {
		// this constructor is made private
		// because i wan't this class to be instatiated by static factory methods
	}

	public HttpRequestDev(byte[] input, byte[] output, ConfigHttpd configHttpd, Config conf) {
		this.configHttpd = configHttpd;
		this.config = conf;
		this.is = new ByteArrayInputStream(input);
		this.os = new ByteArrayOutputStream();
		this.output = output;
	}

	public HttpRequestDev(byte[] input, byte[] output, ConfigHttpd configHttpd, Config conf, InetAddress senderAddress,
			int senderPort) {
		this.configHttpd = configHttpd;
		this.config = conf;
		this.is = new ByteArrayInputStream(input);
		this.os = new ByteArrayOutputStream();
		this.output = output;
		this.senderAddress = senderAddress;
		this.senderPort = senderPort;
	}

	public HttpRequestDev(Socket socket, ConfigHttpd configHttpd, Config conf) throws Exception {
		this.socket = socket;
		this.configHttpd = configHttpd;
		this.config = conf;
	}

	@SuppressWarnings("unused")
	private String contentType(String fileName) {
		StringTokenizer tokens = new StringTokenizer(fileName, ".");
		String currentToken = "";
		while (tokens.hasMoreTokens()) {
			currentToken = tokens.nextToken(); // we are only interested in the
			// last token
		}
		return this.configHttpd.getMimeType(currentToken);
	}

	/**
	 * @return the output
	 */
	public byte[] getOutput() {
		return output;
	}

	/**
	 * this does all the work of processing the http request
	 *
	 * @throws Exception
	 */
	private void processRequest() throws Exception {

		logger.trace("processRequest() - processing Request!");

		InputStreamReader isr = null;
		BufferedReader br = null;

		OutputStreamWriter osw = null;
		BufferedWriter bw = null;

		try {
			configHttpd.countUp();

			if (is == null) {
				is = socket.getInputStream();
			}
			isr = new InputStreamReader(is);
			br = new BufferedReader(isr);
			if (os == null) {
				os = socket.getOutputStream();
				osw = new OutputStreamWriter(os);
				bw = new BufferedWriter(osw);
			}

			// this is the first line containing the relative url and the http method
			String requestLine = crlfReadline(br);
			// To see what's going on we print each HTTP Request (HTTP, HTTPU or HTTPMU) to the log.
			logger.trace("HTTP-REQUEST: " + requestLine);

			// parse all remainig header fields inside the big (and ugly) loop.
			String headerLine = null; // read the request header
			while ((headerLine = crlfReadline(br)).length() != 0) {
				logger.trace("Header Line: " + headerLine);
				if (headerLine.toLowerCase().startsWith("Content-Length:".toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					// String: Content-Length:
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					// the content-length
					if (tokens.hasMoreTokens())
						this.contentLength = Integer.parseInt(tokens.nextToken());
				} else if (headerLine.toLowerCase().startsWith("Host:".toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.host = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(locationStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.location = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(ntStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.nt = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(ntsStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.nts = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(serverStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.server = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(usnStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.usn = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(cacheControlStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.cacheControl = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(stStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.st = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(mxStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.mx = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(manStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.man = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(sidStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.sid = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(subscriptionIdStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.subscriptionId = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(callBackStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.callBack = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(ScopeStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.scope = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(timeOutStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.timeOut = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(optStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.opt = tokens.nextToken();
				} else if (headerLine.toLowerCase().startsWith(o1nlsStr.toLowerCase())) {
					StringTokenizer tokens = new StringTokenizer(headerLine);
					if (tokens.hasMoreTokens())
						tokens.nextToken();
					if (tokens.hasMoreTokens())
						this.o1nls = tokens.nextToken();
				}
			}

			String httpMethod = "";
			String fileName = "";
			String httpVersion = "";

			StringTokenizer tokens = new StringTokenizer(requestLine);
			if (tokens.hasMoreTokens()) {
				httpMethod = tokens.nextToken(); // GET, HEAD, POST, SUBSCRIBE, UNSUBSCRIBE or NOTIFY
			}
			if (tokens.hasMoreTokens()) {
				fileName = tokens.nextToken();
			}
			if (tokens.hasMoreTokens()) {
				httpVersion = tokens.nextToken();
				if (!httpVersion.equalsIgnoreCase("HTTP/1.1")) {
					logger.trace("Wrong HTTP-Version (" + httpVersion + " instead of HTTP/1.1) - ignoring.");
				}
			}

			// build the header
			Header header = new Header(os);

			// Split filename
			StringTokenizer tokens1 = new StringTokenizer(fileName, "?");
			String realFileName = "";
			@SuppressWarnings("unused")
			String arguments = "";
			if (tokens1.hasMoreTokens()) {
				realFileName = tokens1.nextToken();
			}
			if (tokens1.hasMoreTokens()) {
				arguments = tokens1.nextToken();
			}

			String entityBody = Config.CRLF;

			if (httpMethod.equals(ConfigHttpd.HTTP_METHOD_GET) || httpMethod.equals(ConfigHttpd.HTTP_METHOD_POST)) {
				/**
				 * ################################################################
				 * This code handles the HTTPM/U Method GET
				 * ################################################################
				 */
				logger.trace("processRequest() - HTTP_GET/POST");
				if (realFileName.equals("/")) {
					// send the root-xml-file
					logger.trace("HTTP-REQUEST: Serving root-XML");
					header.addHeaderLine(ConfigHttpd.HTTP_STATUS_200);
					header.addHeaderLine("Content-type: text/xml" + "; " + configHttpd.getDefaultCharset());
					entityBody = XMLTools.xmlToString(this.config.getDevice().getXMLElementsWithEnvelope());
					header.addHeaderLine("Content-Length: " + entityBody.length());
					header.sendHeader();
					os.write(entityBody.getBytes());
					logger.trace("GET - Sent Header:\n" + header);
					logger.trace("GET - Sent Body:\n" + entityBody);
				} else if (realFileName.endsWith("_scpd.xml")) {
					// return SCPD-File
					logger.trace("HTTP-REQUEST: Serving SCPD-XML");
					String string = realFileName.substring(2, realFileName.length() - 9);
					String cleanFilename = realFileName.substring(1);
					StringTokenizer tokens2 = new StringTokenizer(string, ".");
					@SuppressWarnings("unused")
					String service = tokens2.nextToken();
					@SuppressWarnings("unused")
					String serviceId = tokens2.nextToken();
					header.addHeaderLine(ConfigHttpd.HTTP_STATUS_200);
					header.addHeaderLine("Content-type: text/xml" + "; " + configHttpd.getDefaultCharset());
					entityBody = XMLTools.xmlToString(this.config.getDevice().getServiceByURL(cleanFilename)
							.getXMLElementsWithEnvelope());
					header.addHeaderLine("Content-Length: " + entityBody.length());
					header.sendHeader();
					os.write(entityBody.getBytes());
					logger.trace("GET - Sent Header:\n" + header);
					logger.trace("GET - Sent Body:\n" + entityBody);
				} else if (realFileName.endsWith("_control")) {
					logger.trace("HTTP-REQUEST: Serving SOAP-Request");
					// This branch handles all incoming SOAP requests
					/*
					 * The SOAP-Request are hard-coded for the most part. This isn't nice, this isn't wise but it's fast
					 * and we don't have much time. At least not this much ...
					 */
					String reqEntityBody = "";
					int buffer = -1;
					while ((reqEntityBody.length() < this.contentLength) && ((buffer = br.read()) != -1)) {
						reqEntityBody += (char) buffer;
					}
					logger.trace("reqEntityBody: \n" + reqEntityBody);
					if (reqEntityBody.length() > 0) {
						Document xml = null;
						SAXBuilder xmlBuilder = new SAXBuilder();
						ByteArrayInputStream bais = new ByteArrayInputStream(reqEntityBody.getBytes());
						xml = xmlBuilder.build(bais);
						logger.trace("Parsed XML: " + XMLTools.xmlToString(xml));
						int loadLevel = 0;
						int powerStatus = 0;
						String getAction = "";
						// ### Handle >> NewLoadLevelTarget <<
						// Try to extract the given field from the XML, if this fails this branch isn't
						// responsible and won't pass the XMLTools call.
						// Then send an appropriate response.
						try {
							loadLevel = Integer.parseInt(XMLTools.getXMLNode(xml.getRootElement().getChildren(),
									"NewLoadLevelTarget").getText());
							this.config.getDevice().getService(Config.DIMMING_SERVICE_ID).setStateVariable(
									"LoadLevelStatus", new Integer(loadLevel).toString());
							entityBody = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n"
									+ "<s:Envelope s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" \r\n"
									+ "\r\n"
									+ "xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\">\r\n"
									+ "   <s:Body>\r\n"
									+ "      <u:SetLoadLevelTargetResponse xmlns:u=\"urn:schemas-upnp-org:service:DimmingService:1\" />\r\n"
									+ "   </s:Body>\r\n" + "</s:Envelope>";
						} catch (NullPointerException e) {
							logger.trace("Got NullPointerException"); // just try to extract the info
						}
						// ### Handle >> newTargetValue <<
						// Try to extract the given field from the XML, if this fails this branch isn't
						// responsible and won't pass the XMLTools call.
						// Then send an appropriate response.
						try {
							powerStatus = Integer.parseInt(XMLTools.getXMLNode(xml.getRootElement().getChildren(),
									"newTargetValue").getText());
							if (powerStatus == 0) {
								this.config.getDevice().getService(Config.SWITCH_POWER_SERVICE_ID).setStateVariable(
										"Status", "0");
							} else {
								this.config.getDevice().getService(Config.SWITCH_POWER_SERVICE_ID).setStateVariable(
										"Status", "1");
							}
							entityBody = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n"
									+ "<s:Envelope s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" \r\n"
									+ "\r\n"
									+ "xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\">\r\n"
									+ "   <s:Body>\r\n"
									+ "      <u:SetTargetResponse xmlns:u=\"urn:schemas-upnp-org:service:SwitchPower:1\" />\r\n"
									+ "   </s:Body>\r\n" + "</s:Envelope>";
						} catch (NullPointerException e) {
							logger.trace("Got NullPointerException", e); // just try to extract the info
						}
						// ### Handle >> GetLoadLevelStatus <<
						// Try to extract the given field from the XML, if this fails this branch isn't
						// responsible and won't pass the XMLTools call.
						// Then send an appropriate response.
						try {
							getAction = XMLTools.getXMLNode(xml.getRootElement().getChildren(), "GetLoadLevelStatus")
									.getName();
							String value = this.config.getDevice().getService(Config.DIMMING_SERVICE_ID)
									.getStateVariableValue("LoadLevelStatus");
							entityBody = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n"
									+ "<s:Envelope s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\">\r\n"
									+ "   <s:Body>\r\n"
									+ "      <u:GetLoadLevelStatusResponse xmlns:u=\"urn:schemas-upnp-org:service:DimmingService:1\">\r\n"
									+ "         <RetLoadLevelStatus>" + value + "</RetLoadLevelStatus>\r\n"
									+ "      </u:GetLoadLevelStatusResponse>\r\n" + "   </s:Body>\r\n"
									+ "</s:Envelope>";
						} catch (NullPointerException e) {
							logger.trace("Got NullPointerException", e); // just try to extract the info
						}
						// ### Handle >> GetMinLevel <<
						// Try to extract the given field from the XML, if this fails this branch isn't
						// responsible and won't pass the XMLTools call.
						// Then send an appropriate response.
						try {
							getAction = XMLTools.getXMLNode(xml.getRootElement().getChildren(), "GetMinLevel")
									.getName();
							String value = this.config.getDevice().getService(Config.DIMMING_SERVICE_ID)
									.getStateVariableValue("MinLevel");
							entityBody = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n"
									+ "<s:Envelope s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\">\r\n"
									+ "   <s:Body>\r\n"
									+ "      <u:GetMinLevelResponse xmlns:u=\"urn:schemas-upnp-org:service:DimmingService:1\">\r\n"
									+ "         <MinLevel>" + value + "</MinLevel>\r\n"
									+ "      </u:GetMinLevelResponse>\r\n" + "   </s:Body>\r\n" + "</s:Envelope>";
						} catch (NullPointerException e) {
							logger.trace("Got NullPointerException", e); // just try to extract the info
						}
						// ### Handle >> GetStatus <<
						// Try to extract the given field from the XML, if this fails this branch isn't
						// responsible and won't pass the XMLTools call.
						// Then send an appropriate response.
						try {
							getAction = XMLTools.getXMLNode(xml.getRootElement().getChildren(), "GetStatus").getName();
							String value = this.config.getDevice().getService(Config.SWITCH_POWER_SERVICE_ID)
									.getStateVariableValue("Status");
							entityBody = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n"
									+ "<s:Envelope s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\">\r\n"
									+ "   <s:Body>\r\n"
									+ "      <u:GetStatusResponse xmlns:u=\"urn:schemas-upnp-org:service:SwitchPower:1\">\r\n"
									+ "         <ResultStatus>" + value + "</ResultStatus>\r\n"
									+ "      </u:GetStatusResponse>\r\n" + "   </s:Body>\r\n" + "</s:Envelope>";
						} catch (NullPointerException e) {
							logger.trace("Got NullPointerException", e); // just try to extract the info
						}
						logger.trace("SOAP-Request, LoadLevel: " + loadLevel + ", PowerStatus: " + powerStatus
								+ ", GetAction: " + getAction);
						// Create a SOAP Response - the entitiy body should have been created aboe.
						header.addHeaderLine(ConfigHttpd.HTTP_STATUS_200);
						header.addHeaderLine("EXT: ");
						header.addHeaderLine("Content-type: text/xml" + "; " + configHttpd.getDefaultCharset());
						header.addHeaderLine("Server: " + Config.getHeaderServer());
						header.addHeaderLine("Content-Length: " + entityBody.length());
						header.sendHeader();
						os.write(entityBody.getBytes());
						logger.trace("GET - Sent Header:\n" + header);
						logger.trace("GET - Sent Body:\n" + entityBody);
					}
				} else if (realFileName.endsWith("_event")) {
					// The Device doesn't handle events. These are for GENA Message in the Control Point only.
					logger.trace("HTTP-REQUEST: Serving Eventing-URL");
					header.addHeaderLine(ConfigHttpd.HTTP_STATUS_501);
					entityBody = "";
					header.addHeaderLine("Content-Length: 0");
					header.sendHeader();
					os.write(entityBody.getBytes());
				} else if (false) {
					// placeholder for a different file
				} else {
					// return a error 404-not found
					// 404 we cant read
					header.addHeaderLine(ConfigHttpd.HTTP_STATUS_404);
					header.addHeaderLine("Content-type: text/html" + "; " + configHttpd.getDefaultCharset());
					entityBody = "<html><head><title>Not Found</title></head>" + "<body><h1>Error 404 - Not Found</h1>"
							+ ConfigHttpd.SIGNATURE + "</body></html>";
					header.addHeaderLine("Content-Length: " + entityBody.length());
					header.sendHeader();
					os.write(entityBody.getBytes());
				}
			} else if (httpMethod.equals(ConfigHttpd.HTTPMU_METHOD_MSEARCH)) {
				/**
				 * ################################################################
				 * This code handles the HTTPM/U Method M-SEARCH
				 * ################################################################
				 */
				logger.trace("processRequest() - HTTP_M-SEARCH");
				header.addHeaderLine(ConfigHttpd.HTTP_STATUS_200);
				header.addHeaderLine("ST: upnp:rootdevice");
				header.addHeaderLine("EXT: ");
				header.addHeaderLine(Config.getHeaderServer());
				header.addHeaderLine("USN: uuid:" + config.getUuid() + "::upnp:rootdevice");
				header.addHeaderLine("CACHE-CONTROL: max-age=3600");
				header.addHeaderLine("LOCATION: http://" + this.config.getLocalAddress().getHostAddress() + ":"
						+ this.config.getHttpServerPort() + "/");
				entityBody = "";
				header.addHeaderLine("Content-Length: 0");
				header.sendHeader();
				logger.trace("Sent Header in reply to M-SEARCH: \n|" + header.toString() + "|");
				logger.trace("Send Body in reply to M-SEARCH: \n|" + entityBody + "|");
				os.write(entityBody.getBytes());
				this.output = ((ByteArrayOutputStream) this.os).toByteArray();
				DiscoverReplyThread drt = new DiscoverReplyThread(this.config, this.senderAddress, this.senderPort, st);
				Thread t = new Thread(drt);
				//Random rand = new Random();
				//int sleep = rand.nextInt(Integer.parseInt(this.mx.trim()));
				//sleep = sleep * 850; // sleep for < (85% of MX)
				//Thread.sleep(sleep);
				Config.waitRandom(this.config, 10, Integer.parseInt(this.mx.trim()) * 1000, true);
				t.start();
			} else if (httpMethod.equals(ConfigHttpd.HTTPMU_METHOD_NOTIFY)) {
				// Since this is a Device we shouldnt listen to NOTIFYs, right?
			} else if (httpMethod.equals(ConfigHttpd.HTTPMU_METHOD_SUBSCRIBE)) {
				/**
				 * ################################################################
				 * This code handles the HTTPM/U Method SUBSCRIBE
				 * ################################################################
				 */
				logger.trace("processRequest() - HTTP_SUBSCRIBE");
				// There SHOULD be no request body for SUBSCRIBE
				logger.trace("SUBSCRIPTION:\nhost: " + this.host + ", callback: " + this.callBack + ", nt: " + nt
						+ ", timeout: " + this.timeOut + ", request: " + requestLine);
				String url = requestLine.substring(11, requestLine.length() - 9);
				logger.trace("Extracted URL: " + url);
				Service sv = this.config.getDevice().getServiceByURL(url);
				String sid = "";
				if (sv == null) {
					logger.error("Couldn't find Service by url: " + url);
				} else {
					logger.trace("Service: " + sv.getControlURL());
					logger.trace("Callback URL: " + this.callBack);
					logger.trace("SID: " + this.sid);
					if (this.callBack.length() > 0) {
						// remove the brackets (< and >) and subscribe
						sid = sv.subscribe(this.callBack.substring(1, this.callBack.length() - 1), this.timeOut);
					} else if (this.sid.length() > 0) {
						sid = this.sid;
						sv.updateTimeOut(sid, this.timeOut);
					} else {
						logger.error("Invalid SID!");
						sid = "-1"; // error :X
					}
				}
				header.addHeaderLine(ConfigHttpd.HTTP_STATUS_200);
				header.addHeaderLine("SID: " + sid);
				header.addHeaderLine(Config.getHeaderServer());
				entityBody = "";
				header.addHeaderLine("Content-Length: 0");
				header.sendHeader();
				os.write(entityBody.getBytes());
			} else if (httpMethod.equals(ConfigHttpd.HTTPMU_METHOD_UNSUBSCRIBE)) {
				/**
				 * ################################################################
				 * This code handles the HTTPM/U Method UNSUBSCRIBE
				 * ################################################################
				 */
				logger.trace("processRequest() - HTTP_UNSUBSCRIBE");
				// There SHOULD be no request body for UNSUBSCRIBE
				logger.trace("UNSUBSCRIPTION:\nhost: " + this.host + ", callback: " + this.callBack + ", nt: " + nt
						+ ", timeout: " + this.timeOut + ", request: " + requestLine + ", sid: " + this.sid);
				String url = requestLine.substring(13, requestLine.length() - 9);
				logger.trace("Extracted URL: " + url);
				Service sv = this.config.getDevice().getServiceByURL(url);
				if (sv == null) {
					logger.error("Couldn't find Service by url: " + url);
				} else {
					logger.trace("Service: " + sv.getControlURL());
					if (sv.unsubscribe(this.sid)) {
						header.addHeaderLine(ConfigHttpd.HTTP_STATUS_200);
						header.addHeaderLine(Config.getHeaderServer());
						entityBody = "";
						header.addHeaderLine("Content-Length: 0");
						header.sendHeader();
						os.write(entityBody.getBytes());
					} else {
						header.addHeaderLine(ConfigHttpd.HTTP_STATUS_412);
						header.addHeaderLine(Config.getHeaderServer());
						entityBody = "";
						header.addHeaderLine("Content-Length: 0");
						header.sendHeader();
						os.write(entityBody.getBytes());
					}
				}

			} else {
				/**
				 * ##############################################################
				 * This code handles every other HTTPM/U
				 * Method (by returning a Code 500 - Internal Server Error)
				 * ##############################################################
				 */
				header.addHeaderLine(ConfigHttpd.HTTP_STATUS_500);
				entityBody = "";
				header.addHeaderLine("Content-Length: 0");
				header.sendHeader();
			}
			os.write(ConfigHttpd.CRLF.getBytes());
		} catch (Exception e) {
			throw e;
		} finally {
			try {
				bw.close();
			} catch (Exception e) {
				logger.trace("processRequest() [bw] - Exception while closing Streams"); // ignore it
			}
			try {
				os.close();
			} catch (Exception e) {
				logger.trace("processRequest() [os] - Exception while closing Streams"); // ignore it
			}
			try {
				br.close();
			} catch (Exception e) {
				logger.trace("processRequest() [br] - Exception while closing Streams"); // ignore it
			}
			try {
				osw.close();
			} catch (Exception e) {
				logger.trace("processRequest() [osw] - Exception while closing Streams"); // ignore it
			}
			try {
				this.output = ((ByteArrayOutputStream) this.os).toByteArray();
			} catch (Exception e) {
				logger.trace("processRequest() [boas.toByteArray] - Exception while closing Streams"); // ignore it
			}
			configHttpd.countDown();
		}
	}

	/**
	 * runs the http-Request-Thread
	 */
	public void run() {
		try {
			processRequest();
		} catch (Exception e) {
			logger.trace(e);
		}
	}
}
