package device.structures;

import java.net.URL;
import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.jdom.Element;
import org.jdom.Namespace;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import common.core.Config;
import common.structures.Icon;
import common.structures.Service;
import device.multicast.SSDPAliveThread;
import device.multicast.SSDPByeByeThread;

/**
 * <pre>
 *          Copyright (c) 2007 Dominik Schulz
 *          Copyright (c) 2007 Florian Lindner
 *          Copyright (c) 2007 Philip Hartmann
 *
 *          This file is part of jUPnP.
 *
 *          jUPnP is free software; you can redistribute it and/or modify
 *          it under the terms of the GNU General Public License as published by
 *          the Free Software Foundation; either version 2 of the License, or
 *          (at your option) any later version.
 *
 *          jUPnP is distributed in the hope that it will be useful,
 *          but WITHOUT ANY WARRANTY; without even the implied warranty of
 *          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *          GNU General Public License for more details.
 *
 *          You should have received a copy of the GNU General Public License
 *          along with jUPnP; if not, write to the Free Software
 *          Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * </pre>
 *
 * @author tex
 *
 * "According to Einstein's theory of relativity, Chuck Norris can actually roundhouse kick you yesterday."
 */
public class Device {

	static Logger				logger				= Logger.getLogger(Device.class);

	private String				deviceType			= "";

	private String				friendlyName		= "";

	private String				manufacturer		= "";

	private URL					manufacturerURL		= null;

	private String				modelDescription	= "";

	private String				modelName			= "";

	private String				modelNumber			= "";

	private URL					modelURL			= null;

	private String				serialNumber		= "";

	private String				UDN					= "";

	private String				UPC					= "";

	private String				urlBase				= "";

	private URL					presentationURL		= null;

	private URL					descriptionURL		= null;

	private boolean				isRootdevice		= false;

	private ArrayList<Device>	embeddedDevices		= new ArrayList<Device>();

	private ArrayList<Service>	services			= new ArrayList<Service>();

	private ArrayList<Icon>		icons				= new ArrayList<Icon>();

	private Config				config				= null;

	/**
	 * A Device can inlude several devices. This getter returns the one with the given name
	 *
	 * @param name
	 * @return
	 */
	public Device getEmbeddedDevice(String name) {
		for (int i = 0; i < this.embeddedDevices.size(); i++) {
			if (this.embeddedDevices.get(i).deviceType.equalsIgnoreCase(name)) {
				return this.embeddedDevices.get(i);
			}
		}
		return new Device(false, this.config);
	}

	/**
	 *
	 * @param name
	 * @return
	 */
	public Service getService(String name) {
		logger.trace("getService(" + name + ") - Number of Services: " + this.services.size());
		for (int i = 0; i < this.services.size(); i++) {
			logger.trace("getService(" + name + ") - Comparing to service: " + this.services.get(i).getServiceID());
			if (this.services.get(i).getServiceID().equals(name)) {
				logger.trace("Found correct Service!");
				return this.services.get(i);
			}
		}
		logger.error("Service named '" + name + "' not found. Returning arbitrary one.");
		return new Service(this.config);
	}

	/**
	 * informs all subscribers of all services about all state vars
	 *
	 */
	public void notifyChangeAll() {
		for (int i = 0; i < this.services.size(); i++) {
			this.services.get(i).notifyChangeAll();
		}
	}

	/**
	 * Tell all sevices to check for expired subscriptions.
	 *
	 */
	public void subscriptionsCleanUp() {
		for (int i = 0; i < this.services.size(); i++) {
			this.services.get(i).cleanUp();
		}
	}

	/**
	 *
	 * @param isRootdevice
	 *            if this should be a root device or not (for embedded devices)
	 * @param conf
	 *            Our Config
	 */
	public Device(boolean isRootdevice, Config conf) {
		this.isRootdevice = isRootdevice;
		this.config = conf;
	}

	/**
	 * Add an embedded device
	 *
	 * @param dev
	 * @return
	 */
	public boolean addEmbeddedDevice(Device dev) {
		return this.embeddedDevices.add(dev);
	}

	/**
	 *
	 * @param svc
	 * @return
	 */
	public boolean addService(Service svc) {
		return this.services.add(svc);
	}

	/**
	 *
	 * @return
	 */
	public Config getConfig() {
		return config;
	}

	/**
	 * @return the descriptionURL
	 */
	public URL getDescriptionURL() {
		return descriptionURL;
	}

	/**
	 * @return the deviceType
	 */
	public String getDeviceType() {
		return this.deviceType;
	}

	/**
	 * @return the embeddedDevices
	 */
	public ArrayList<Device> getEmbeddedDevices() {
		return embeddedDevices;
	}

	/**
	 * @return the friendlyName
	 */
	public String getFriendlyName() {
		return friendlyName;
	}

	/**
	 * @return the icons
	 */
	public ArrayList<Icon> getIcons() {
		return icons;
	}

	/**
	 * @return the manufacturer
	 */
	public String getManufacturer() {
		return manufacturer;
	}

	/**
	 * @return the manufacturerURL
	 */
	public URL getManufacturerURL() {
		return manufacturerURL;
	}

	/**
	 * @return the modelDescription
	 */
	public String getModelDescription() {
		return modelDescription;
	}

	/**
	 * @return the modelName
	 */
	public String getModelName() {
		return modelName;
	}

	/**
	 * @return the modelNumber
	 */
	public String getModelNumber() {
		return modelNumber;
	}

	/**
	 * @return the modelURL
	 */
	public URL getModelURL() {
		return modelURL;
	}

	/**
	 * @return the presentationURL
	 */
	public URL getPresentationURL() {
		return presentationURL;
	}

	/**
	 *
	 * @return the Real Device Type
	 */
	public String getRealDeviceType() {
		return deviceType;
	}

	/**
	 * @return the serialNumber
	 */
	public String getSerialNumber() {
		return serialNumber;
	}

	/**
	 * Selects the correct service for the given URL.
	 *
	 * @param url
	 * @return
	 */
	public Service getServiceByURL(String url) {
		logger.trace("getServiceByURL(" + url + ") - comparing to: ");
		for (int i = 0; i < this.getServices().size(); i++) {
			logger.trace("SCPDURL: " + this.getServices().get(i).getDescriptionURL());
			logger.trace("ControlURL: " + this.getServices().get(i).getControlURL());
			logger.trace("EventURL: " + this.getServices().get(i).getEventSubURL());
			if (this.getServices().get(i).getControlURL().equals(url)) {
				return this.getServices().get(i);
			}
			if (this.getServices().get(i).getEventSubURL().equals(url)) {
				return this.getServices().get(i);
			}
			if (this.getServices().get(i).getDescriptionURL().equals(url)) {
				return this.getServices().get(i);
			}
		}
		logger.error("Didn't find any matching service!");
		return null;
	}

	/**
	 * @return the services
	 */
	public ArrayList<Service> getServices() {
		return services;
	}

	/**
	 * @return the uDN
	 */
	public String getUDN() {
		return UDN;
	}

	/**
	 * @return the uPC
	 */
	public String getUPC() {
		return UPC;
	}

	/**
	 * @return the urlBase
	 */
	public String getUrlBase() {
		return urlBase;
	}

	/**
	 *
	 * @return the URN
	 */
	public String getURN() {
		return this.getDeviceType();
	}

	/**
	 *
	 * @return the UUID String
	 */
	public String getUUID() {
		return "uuid:" + this.config.getUuid();
	}

	/**
	 *
	 * @return a XML-Structure (Element) representing this node
	 */
	public Element getXMLElements() {
		Element root = new Element("device");

		Element deviceType = new Element("deviceType");
		deviceType.setText(this.getDeviceType());
		root.addContent(deviceType);

		Element friendlyName = new Element("friendlyName");
		friendlyName.setText(this.friendlyName);
		root.addContent(friendlyName);

		Element manufacturer = new Element("manufacturer");
		manufacturer.setText(this.manufacturer);
		root.addContent(manufacturer);

		Element manufacturerURL = new Element("manufacturerURL");
		if (this.manufacturerURL != null)
			manufacturerURL.setText(this.manufacturerURL.toString());
		root.addContent(manufacturerURL);

		Element modelDescription = new Element("modelDescription");
		modelDescription.setText(this.modelDescription);
		root.addContent(modelDescription);

		Element modelName = new Element("modelName");
		modelName.setText(this.modelName);
		root.addContent(modelName);

		Element modelNumber = new Element("modelNumber");
		modelNumber.setText(this.modelNumber);
		root.addContent(modelNumber);

		Element modelURL = new Element("modelURL");
		if (this.modelURL != null)
			modelURL.setText(this.modelURL.toString());
		root.addContent(modelURL);

		Element serialNumber = new Element("serialNumber");
		serialNumber.setText(this.serialNumber);

		Element UDN = new Element("UDN");
		UDN.setText(this.UDN);
		root.addContent(UDN);

		Element UPC = new Element("UPC");
		UPC.setText(this.UPC);

		Element urlBase = new Element("UrlBase");
		urlBase.setText(this.urlBase);

		Element presentationURL = new Element("presentationURL");
		if (this.presentationURL != null)
			presentationURL.setText(this.presentationURL.toString());

		Element descriptionURL = new Element("descriptionURL");
		if (this.descriptionURL != null)
			descriptionURL.setText(this.descriptionURL.toString());

		Element embeddedDevices = new Element("embeddedDevices");
		for (int i = 0, n = this.embeddedDevices.size(); i < n; i++) {
			embeddedDevices.addContent(this.embeddedDevices.get(i).getXMLElements());
		}

		Element services = new Element("serviceList");
		for (int i = 0, n = this.services.size(); i < n; i++) {
			services.addContent(this.services.get(i).getXMLElementsForDevice());
		}
		root.addContent(services);

		Element icons = new Element("iconList");
		for (int i = 0, n = this.icons.size(); i < n; i++) {
			icons.addContent(this.icons.get(i).getXMLElements());
		}

		return root;
	}

	/**
	 *
	 * @return The relevant sub-nodes of the device node inside a root element as necessary for the XML-Device
	 *         description.
	 */
	public Element getXMLElementsWithEnvelope() {
		Element root = new Element("root");
		Namespace ns = Namespace.getNamespace("urn:schemas-upnp-org:device-1-0");
		root.setNamespace(ns);
		Element specVersion = new Element("specVersion");
		Element major = new Element("major");
		major.setText("1");
		Element minor = new Element("minor");
		minor.setText("0");
		specVersion.addContent(major);
		specVersion.addContent(minor);
		root.addContent(specVersion);
		root.addContent(this.getXMLElements());
		return root;
	}

	/**
	 *
	 * @return a pretty formated XML String representing this node
	 */
	public String getXMLString() {
		Element root = new Element("root");
		Namespace ns = Namespace.getNamespace("urn:schemas-upnp-org:device-1-0");
		root.setNamespace(ns);
		Element specVersion = new Element("specVersion");
		Element major = new Element("major");
		major.setText("1");
		Element minor = new Element("minor");
		minor.setText("0");
		specVersion.addContent(major);
		specVersion.addContent(minor);
		root.addContent(specVersion);
		root.addContent(this.getXMLElements());
		XMLOutputter xmlOut = new XMLOutputter();
		xmlOut.setFormat(Format.getPrettyFormat());
		return xmlOut.outputString(root);
	}

	/**
	 * @return the isRootdevice
	 */
	public boolean isRootdevice() {
		return isRootdevice;
	}

	/**
	 * Notify about this devices presence.
	 *
	 */
	public void sendSSDPAlive() {
		this.sendSSDPAlive(true, false);
	}

	/**
	 * this device tells about its presence on the network.
	 *
	 * @param descend
	 *            Descend into child strctures or not (for e.g. ST: ssdp:rootdevice)
	 * @param rootdevonly
	 *            Only send notify for the rootdevice or for all devices
	 */
	public void sendSSDPAlive(boolean descend, boolean rootdevonly) {
		SSDPAliveThread ssdpat = null;
		if (this.isRootdevice()) {
			// send upnp:rootdevice message
			ssdpat = new SSDPAliveThread(this.config, "upnp:rootdevice", this.getUUID() + "::upnp:rootdevice");
			Thread t = new Thread(ssdpat);
			t.start();
			if (!rootdevonly) {
				// send upnp:device message
				ssdpat = new SSDPAliveThread(this.config, this.getURN(), this.getUUID() + "::" + this.getURN());
				t = new Thread(ssdpat);
				t.start();
				// send uuid-message
				ssdpat = new SSDPAliveThread(this.config, this.getUUID(), this.getUUID());
				t = new Thread(ssdpat);
				t.start();
			}
		} else {
			ssdpat = new SSDPAliveThread(this.config, this.getURN(), this.getUUID() + "::" + this.getURN());
			Thread t = new Thread(ssdpat);
			t.start();
		}
		if (descend) {
			for (int i = 0; i < services.size(); i++) {
				services.get(i).sendSSDPAlive();
			}
			for (int i = 0; i < embeddedDevices.size(); i++) {
				embeddedDevices.get(i).sendSSDPAlive(descend, rootdevonly);
			}
		}
	}

	/**
	 * the device tells that it leaves the network
	 *
	 */
	public void sendSSDPByeBye() {
		SSDPByeByeThread ssdpat = null;
		if (this.isRootdevice()) {
			// send upnp:rootdevice message
			ssdpat = new SSDPByeByeThread(this.config, "upnp:rootdevice", this.getUUID() + "::upnp:rootdevice");
			Thread t = new Thread(ssdpat);
			t.start();
			// send upnp:device message
			ssdpat = new SSDPByeByeThread(this.config, this.getURN(), this.getUUID() + "::" + this.getURN());
			t = new Thread(ssdpat);
			t.start();
			// send uuid-message
			ssdpat = new SSDPByeByeThread(this.config, this.getUUID(), this.getUUID());
			t = new Thread(ssdpat);
			t.start();
		} else {
			ssdpat = new SSDPByeByeThread(this.config, this.getURN(), this.getUUID() + "::" + this.getURN());
			Thread t = new Thread(ssdpat);
			t.start();
		}
		//Send byebye for all Services
		for (int i = 0; i < services.size(); i++) {
			services.get(i).sendSSDPByeBye();
		}
		//also send byebye for the embedded devices
		for (int i = 0; i < embeddedDevices.size(); i++) {
			embeddedDevices.get(i).sendSSDPByeBye();
		}
	}

	/**
	 * @param descriptionURL
	 *            the descriptionURL to set
	 */
	public void setDescriptionURL(URL descriptionURL) {
		this.descriptionURL = descriptionURL;
	}

	/**
	 * @param deviceType
	 *            the deviceType to set
	 */
	public void setDeviceType(String deviceType) {
		this.deviceType = deviceType;
	}

	/**
	 * @param embeddedDevices
	 *            the embeddedDevices to set
	 */
	public void setEmbeddedDevices(ArrayList<Device> embeddedDevices) {
		this.embeddedDevices = embeddedDevices;
	}

	/**
	 * @param friendlyName
	 *            the friendlyName to set
	 */
	public void setFriendlyName(String friendlyName) {
		this.friendlyName = friendlyName;
	}

	/**
	 * @param icons
	 *            the icons to set
	 */
	public void setIcons(ArrayList<Icon> icons) {
		this.icons = icons;
	}

	/**
	 * @param manufacturer
	 *            the manufacturer to set
	 */
	public void setManufacturer(String manufacturer) {
		this.manufacturer = manufacturer;
	}

	/**
	 * @param manufacturerURL
	 *            the manufacturerURL to set
	 */
	public void setManufacturerURL(URL manufacturerURL) {
		this.manufacturerURL = manufacturerURL;
	}

	/**
	 * @param modelDescription
	 *            the modelDescription to set
	 */
	public void setModelDescription(String modelDescription) {
		this.modelDescription = modelDescription;
	}

	/**
	 * @param modelName
	 *            the modelName to set
	 */
	public void setModelName(String modelName) {
		this.modelName = modelName;
	}

	/**
	 * @param modelNumber
	 *            the modelNumber to set
	 */
	public void setModelNumber(String modelNumber) {
		this.modelNumber = modelNumber;
	}

	/**
	 * @param modelURL
	 *            the modelURL to set
	 */
	public void setModelURL(URL modelURL) {
		this.modelURL = modelURL;
	}

	/**
	 * @param presentationURL
	 *            the presentationURL to set
	 */
	public void setPresentationURL(URL presentationURL) {
		this.presentationURL = presentationURL;
	}

	/**
	 * @param serialNumber
	 *            the serialNumber to set
	 */
	public void setSerialNumber(String serialNumber) {
		this.serialNumber = serialNumber;
	}

	/**
	 * @param services
	 *            the services to set
	 */
	public void setServices(ArrayList<Service> services) {
		this.services = services;
	}

	/**
	 * @param udn
	 *            the uDN to set
	 */
	public void setUDN(String udn) {
		UDN = udn;
	}

	/**
	 * @param upc
	 *            the uPC to set
	 */
	public void setUPC(String upc) {
		UPC = upc;
	}

	/**
	 * @param urlBase
	 *            the urlBase to set
	 */
	public void setUrlBase(String urlBase) {
		this.urlBase = urlBase;
	}
}
